/*
 * GNF.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef GNF_H_
#define GNF_H_

#include <map>
#include <vector>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../GrammarBase.h"
#include "../GrammarFeatures.h"
#include "../GrammarException.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../common/GrammarNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

#include <core/normalize.hpp>

namespace grammar {

/**
 * Context free grammar in greibach normal form. Type 2 in Chomsky hierarchy. Produces context free languages.
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

template < class SymbolType >
class GNF final : public GrammarBase, public alib::Components < GNF < SymbolType >, SymbolType, ext::tuple < TerminalAlphabet, NonterminalAlphabet >, ext::tuple < InitialSymbol > > {
	ext::map < SymbolType, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > > rules;
	bool generatesEpsilon;

public:
	explicit GNF ( SymbolType initialSymbol );

	explicit GNF ( ext::set < SymbolType > nonTerminalSymbols, ext::set < SymbolType > terminalSymbols, SymbolType initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( SymbolType leftHandSide, ext::pair < SymbolType, ext::vector < SymbolType > > rightHandSide );

	void addRules ( SymbolType leftHandSide, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > rightHandSide );

	const ext::map < SymbolType, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > > & getRules ( ) const &;

	ext::map < SymbolType, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > > && getRules ( ) &&;

	bool removeRule ( const SymbolType & leftHandSide, const ext::pair < SymbolType, ext::vector < SymbolType > > & rightHandSide );

	bool addRawRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide );

	ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > getRawRules ( ) const;

	bool removeRawRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide );

	const SymbolType & getInitialSymbol ( ) const & {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	SymbolType && getInitialSymbol ( ) && {
		return std::move ( this->template accessElement < InitialSymbol > ( ).get ( ) );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const ext::set < SymbolType > & getNonterminalAlphabet ( ) const & {
		return this->template accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getNonterminalAlphabet ( ) && {
		return std::move ( this->template accessComponent < NonterminalAlphabet > ( ).get ( ) );
	}

	bool addNonterminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const ext::set < SymbolType > & getTerminalAlphabet ( ) const & {
		return this->template accessComponent < TerminalAlphabet > ( ).get ( );
	}

	ext::set < SymbolType > && getTerminalAlphabet ( ) && {
		return std::move ( this->template accessComponent < TerminalAlphabet > ( ).get ( ) );
	}

	bool addTerminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void setGeneratesEpsilon ( bool genEps );

	bool getGeneratesEpsilon ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const GNF & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "GNF";

		return xmlTagName;
	}

	static GNF parse ( ext::deque < sax::Token >::iterator & input );
	static void parseRule ( ext::deque < sax::Token >::iterator & input, GNF & grammar );

	static void compose ( ext::deque < sax::Token > & out, const GNF & grammar );
	static void composeRules ( ext::deque < sax::Token > & out, const GNF & grammar );

	virtual alib::ObjectBase * inc ( ) &&;

	typedef GNF < > normalized_type;
};

template < class SymbolType >
GNF < SymbolType >::GNF ( SymbolType initialSymbol ) : GNF ( ext::set < SymbolType > { initialSymbol }, ext::set < SymbolType > ( ), initialSymbol ) {
}

template < class SymbolType >
GNF < SymbolType >::GNF ( ext::set < SymbolType > nonterminalAlphabet, ext::set < SymbolType > terminalAlphabet, SymbolType initialSymbol ) : alib::Components < GNF, SymbolType, ext::tuple < TerminalAlphabet, NonterminalAlphabet >, ext::tuple < InitialSymbol > > ( ext::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), ext::make_tuple ( std::move ( initialSymbol ) ) ), generatesEpsilon ( false ) {
}

template < class SymbolType >
GrammarBase * GNF < SymbolType >::clone ( ) const {
	return new GNF ( * this );
}

template < class SymbolType >
GrammarBase * GNF < SymbolType >::plunder ( ) && {
	return new GNF ( std::move ( * this ) );
}

template < class SymbolType >
bool GNF < SymbolType >::addRule ( SymbolType leftHandSide, ext::pair < SymbolType, ext::vector < SymbolType > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	if ( !getTerminalAlphabet ( ).count ( rightHandSide.first ) )
		throw GrammarException ( "Rule must rewrite to terminal symbol" );

	for ( const SymbolType & rhsNTs : rightHandSide.second )
		if ( !getNonterminalAlphabet ( ).count ( rhsNTs ) )
			throw GrammarException ( "Symbol \"" + ext::to_string ( rhsNTs ) + "\" is not a nonterminal symbol" );

	return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
}

template < class SymbolType >
void GNF < SymbolType >::addRules ( SymbolType leftHandSide, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const std::pair < SymbolType, ext::vector < SymbolType > > & rhs : rightHandSide ) {
		if ( ! getTerminalAlphabet ( ).count ( rhs.first ) )
			throw GrammarException ( "Rule must rewrite to terminal symbol" );

		for ( const SymbolType & rhsNTs : rhs.second )
			if ( !getNonterminalAlphabet ( ).count ( rhsNTs ) )
				throw GrammarException ( "Symbol \"" + ext::to_string ( rhsNTs ) + "\" is not a nonterminal symbol" );
	}

	rules[std::move ( leftHandSide )].insert ( ext::make_moveable_set ( rightHandSide ).begin ( ), ext::make_moveable_set ( rightHandSide ).end ( ) );
}

template < class SymbolType >
const ext::map < SymbolType, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > > & GNF < SymbolType >::getRules ( ) const & {
	return rules;
}

template < class SymbolType >
ext::map < SymbolType, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > > && GNF < SymbolType >::getRules ( ) && {
	return std::move ( rules );
}

template < class SymbolType >
bool GNF < SymbolType >::removeRule ( const SymbolType & leftHandSide, const ext::pair < SymbolType, ext::vector < SymbolType > > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

template < class SymbolType >
bool GNF < SymbolType >::addRawRule ( SymbolType leftHandSide, ext::vector < SymbolType > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( true );
		return !res;
	} else {
		SymbolType first = std::move ( rightHandSide[0] );
		ext::vector < SymbolType > rest ( std::make_move_iterator ( rightHandSide.begin ( ) ) + 1, std::make_move_iterator ( rightHandSide.end ( ) ) );
		return addRule ( std::move ( leftHandSide ), ext::make_pair ( std::move ( first ), std::move ( rest ) ) );
	}
}

template < class SymbolType >
ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > GNF < SymbolType >::getRawRules ( ) const {
	ext::map < SymbolType, ext::set < ext::vector < SymbolType > > > res;

	for ( const auto & rule : getRules ( ) )
		for ( const auto & rhs : rule.second ) {
			ext::vector < SymbolType > tmp { rhs.first };
			tmp.insert ( tmp.end ( ), rhs.second.begin ( ), rhs.second.end ( ) );
			res[rule.first].insert ( std::move ( tmp ) );
		}

	if ( getGeneratesEpsilon ( ) )
		res[getInitialSymbol ( )].insert ( ext::vector < SymbolType > { } );

	return res;
}

template < class SymbolType >
bool GNF < SymbolType >::removeRawRule ( const SymbolType & leftHandSide, const ext::vector < SymbolType > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( false );
		return res;
	} else {
		SymbolType first = rightHandSide[0];
		ext::vector < SymbolType > rest ( rightHandSide.begin ( ) + 1, rightHandSide.end ( ) );
		return removeRule ( leftHandSide, ext::make_pair ( first, rest ) );
	}
}

template < class SymbolType >
void GNF < SymbolType >::setGeneratesEpsilon ( bool genEps ) {
	generatesEpsilon = genEps;
}

template < class SymbolType >
bool GNF < SymbolType >::getGeneratesEpsilon ( ) const {
	return generatesEpsilon;
}

template < class SymbolType >
int GNF < SymbolType >::compare ( const GNF & other ) const {
	auto first = ext::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = ext::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void GNF < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(GNF "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules << ")";
}

template < class SymbolType >
GNF < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
GNF < SymbolType > GNF < SymbolType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, GNF::getXmlTagName() );

	ext::set < SymbolType > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet < SymbolType > ( input );
	ext::set < SymbolType > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet < SymbolType > ( input );
	SymbolType initialSymbol = GrammarFromXMLParser::parseInitialSymbol < SymbolType > ( input );

	GNF < SymbolType > grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon ( input );
	grammar.setGeneratesEpsilon ( generatesEpsilon );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, GNF::getXmlTagName() );
	return grammar;
}

template < class SymbolType >
void GNF < SymbolType >::parseRule ( ext::deque < sax::Token >::iterator & input, GNF & grammar ) {
	SymbolType lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS < SymbolType > ( input );
	ext::pair < SymbolType, ext::vector < SymbolType > > rhs = GrammarFromXMLParser::parseRuleGNFRHS < SymbolType > ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

template < class SymbolType >
void GNF < SymbolType >::compose ( ext::deque < sax::Token > & out, const GNF & grammar ) {
	out.emplace_back ( GNF::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, grammar.getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, grammar.getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, grammar.getInitialSymbol ( ) );
	composeRules ( out, grammar );
	GrammarToXMLComposer::composeGeneratesEpsilon ( out, grammar.getGeneratesEpsilon ( ) );

	out.emplace_back ( GNF::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
void GNF < SymbolType >::composeRules ( ext::deque < sax::Token > & out, const GNF & grammar ) {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : grammar.getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleGNFRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* GNF < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace grammar */

namespace alib {

template < class SymbolType >
class ComponentConstraint< grammar::GNF < SymbolType >, SymbolType, grammar::TerminalAlphabet > {
public:
	static bool used ( const grammar::GNF < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > > & rule : grammar.getRules ( ) )
			for ( const std::pair < SymbolType, ext::vector < SymbolType > > & rhs : rule.second )
				if ( rhs.first == symbol )
					return true;

		return false;
	}

	static bool available ( const grammar::GNF < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::GNF < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + ext::to_string ( symbol ) + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
	}
};

template < class SymbolType >
class ComponentConstraint< grammar::GNF < SymbolType >, SymbolType, grammar::NonterminalAlphabet > {
public:
	static bool used ( const grammar::GNF < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > > & rule : grammar.getRules ( ) ) {
			if ( rule.first == symbol )
				return true;

			for ( const std::pair < SymbolType, ext::vector < SymbolType > > & rhs : rule.second )
				if ( std::find ( rhs.second.begin ( ), rhs.second.end ( ), symbol ) != rhs.second.end ( ) )
					return true;
		}

		if ( grammar.template accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
			return true;

		return false;
	}

	static bool available ( const grammar::GNF < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::GNF < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + ext::to_string ( symbol ) + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
	}
};

template < class SymbolType >
class ElementConstraint< grammar::GNF < SymbolType >, SymbolType, grammar::InitialSymbol > {
public:
	static bool available ( const grammar::GNF < SymbolType > & grammar, const SymbolType & symbol ) {
		return grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const grammar::GNF < SymbolType > &, const SymbolType & ) {
	}
};

template < class SymbolType >
struct normalize < grammar::GNF < SymbolType >, typename std::enable_if < ! std::is_same < grammar::GNF < SymbolType >, grammar::GNF < > >::value >::type > {
	static grammar::GNF < > eval ( grammar::GNF < SymbolType > && value ) {
		ext::set < DefaultSymbolType > nonterminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getNonterminalAlphabet ( ) );
		ext::set < DefaultSymbolType > terminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getTerminalAlphabet ( ) );
		DefaultSymbolType initialSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( value ).getInitialSymbol ( ) );

		grammar::GNF < > res ( std::move ( nonterminals ), std::move ( terminals ), std::move ( initialSymbol ) );

		for ( std::pair < SymbolType, ext::set < ext::pair < SymbolType, ext::vector < SymbolType > > > > && rule : ext::make_moveable_map ( std::move ( value ).getRules ( ) ) ) {

			ext::set < ext::pair < DefaultSymbolType, ext::vector < DefaultSymbolType > > > rhs;
			for ( ext::pair < SymbolType, ext::vector < SymbolType > > && target : ext::make_moveable_set ( rule.second ) )
				rhs.insert ( grammar::GrammarNormalize::normalizeRHS ( std::move ( target ) ) );

			DefaultSymbolType lhs = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( rule.first ) );

			res.addRules ( std::move ( lhs ), std::move ( rhs ) );
		}

		res.setGeneratesEpsilon ( value.getGeneratesEpsilon ( ) );

		return res;
	}
};

} /* namespace alib */

#endif /* GNF_H_ */
