/*
 * NonlinearCompressedBitParallelTreeIndex.h
 *
 *  Created on: 22. 8. 2017
 *      Author: Jan Travnicek
 */

#ifndef ARBOLOGY_NONLINEAR_COMPRESSED_BIT_PARALLEL_INDEX_H_
#define ARBOLOGY_NONLINEAR_COMPRESSED_BIT_PARALLEL_INDEX_H_

#include <string>
#include <sstream>

#include <alphabet/RankedSymbol.h>
#include <alphabet/xml/RankedSymbol.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsMap.h>
#include <container/ObjectsVector.h>
#include <container/xml/ObjectsSet.h>
#include <container/xml/ObjectsMap.h>
#include <container/xml/ObjectsVector.h>
#include <common/SparseBoolVector.hpp>
#include <primitive/Bool.h>
#include <primitive/xml/Bool.h>
#include <primitive/xml/Integer.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace arbology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType, class RankType = DefaultRankType >
class NonlinearCompressedBitParallelTreeIndex final : public alib::ObjectBase, public alib::Components < NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ext::tuple < GeneralAlphabet >, ext::tuple < > > {
protected:
	ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > m_vectors;
	ext::vector < int > m_jumpTable;
	ext::vector < unsigned > m_repeats;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit NonlinearCompressedBitParallelTreeIndex ( ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet, ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > vectors, ext::vector < int > jumpTable, ext::vector < unsigned > repeats );

	/**
	 * @return Root node of the trie
	 */
	const ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > & getData ( ) const &;

	ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > && getData ( ) &&;

	/**
	 * @return subtree jump table
	 */
	const ext::vector < int > & getJumps ( ) const &;

	ext::vector < int > && getJumps ( ) &&;

	const ext::vector < unsigned > & getRepeats ( ) const &;

	ext::vector < unsigned > && getRepeats ( ) &&;

	ext::vector < common::ranked_symbol < SymbolType, RankType > > getString ( ) const;

	const ext::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < common::ranked_symbol < SymbolType, RankType > > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	/**
	 * Sets the nonlinearcompressedBit vector for given symbol
	 * @param tree root node to set
	 */
	void setNonlinearCompressedBitVectorForSymbol ( common::ranked_symbol < SymbolType, RankType > symbol, common::SparseBoolVector data );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NonlinearCompressedBitParallelTreeIndex & other ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef NonlinearCompressedBitParallelTreeIndex < > normalized_type;
};

} /* namespace arbology */

} /* namespace indexes */

namespace indexes {

namespace arbology {

template < class SymbolType, class RankType >
NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::NonlinearCompressedBitParallelTreeIndex ( ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet, ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > vectors, ext::vector < int > jumpTable, ext::vector < unsigned > repeats ) : alib::Components < NonlinearCompressedBitParallelTreeIndex, common::ranked_symbol < SymbolType, RankType >, ext::tuple < GeneralAlphabet >, ext::tuple < > > ( ext::make_tuple ( std::move ( alphabet ) ), ext::tuple < > ( ) ), m_vectors ( std::move ( vectors ) ), m_jumpTable ( std::move ( jumpTable ) ), m_repeats ( std::move ( repeats ) ) {
}

template < class SymbolType, class RankType >
alib::ObjectBase * NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::clone ( ) const {
	return new NonlinearCompressedBitParallelTreeIndex ( * this );
}

template < class SymbolType, class RankType >
alib::ObjectBase * NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::plunder ( ) && {
	return new NonlinearCompressedBitParallelTreeIndex ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > & NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::getData ( ) const & {
	return m_vectors;
}

template < class SymbolType, class RankType >
ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > && NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::getData ( ) && {
	return std::move ( m_vectors );
}

template < class SymbolType, class RankType >
const ext::vector < int > & NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::getJumps ( ) const & {
	return m_jumpTable;
}

template < class SymbolType, class RankType >
ext::vector < int > && NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::getJumps ( ) && {
	return std::move ( m_jumpTable );
}

template < class SymbolType, class RankType >
const ext::vector < unsigned > & NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::getRepeats ( ) const & {
	return m_repeats;
}

template < class SymbolType, class RankType >
ext::vector < unsigned > && NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::getRepeats ( ) && {
	return std::move ( m_repeats );
}

template < class SymbolType, class RankType >
ext::vector < common::ranked_symbol < SymbolType, RankType > > NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::getString ( ) const {
	ext::vector < common::ranked_symbol < SymbolType, RankType > > res;

	unsigned index = 0;

	do {
		for ( const std::pair < const common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > & nonlinearcompressedBitVector : m_vectors )
			if ( nonlinearcompressedBitVector.second.size ( ) > index && nonlinearcompressedBitVector.second [ index ] ) {
				res.push_back ( nonlinearcompressedBitVector.first );
				continue;
			}

	} while ( res.size ( ) == index ++ + 1 );

	return res;
}

template < class SymbolType, class RankType >
void NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::setNonlinearCompressedBitVectorForSymbol ( common::ranked_symbol < SymbolType, RankType > symbol, common::SparseBoolVector data ) {
	this->m_vectors [ symbol ] = std::move ( data );
}

template < class SymbolType, class RankType >
void NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(NonlinearCompressedBitParallelTreeIndex " << this->m_vectors << ", " << m_jumpTable << ")";
}

template < class SymbolType, class RankType >
int NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::compare ( const NonlinearCompressedBitParallelTreeIndex & other ) const {
	auto first = ext::tie ( getData ( ), getAlphabet ( ), getJumps ( ), getRepeats ( ) );
	auto second = ext::tie ( other.getData ( ), other.getAlphabet ( ), other.getJumps ( ), getRepeats ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class RankType >
alib::ObjectBase * NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >::inc ( ) && {
	return new alib::UniqueObject ( alib::Object ( std::move ( * this ) ), primitive::Integer ( 0 ) );
}

} /* namespace arbology */

} /* namespace indexes */

namespace alib {

template < class SymbolType, class RankType >
class ComponentConstraint < indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, indexes::arbology::GeneralAlphabet > {
public:
	static bool used ( const indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > & index, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		const ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > & content = index.getData ( );

		return content.find ( symbol ) != content.end ( );
	}

	static bool available ( const indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
	}

};

template < class SymbolType, class RankType >
struct normalize < indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >, typename std::enable_if < ! std::is_same < indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType >, indexes::arbology::NonlinearCompressedBitParallelTreeIndex < > >::value >::type > {
	static indexes::arbology::NonlinearCompressedBitParallelTreeIndex < > eval ( indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > && value ) {
		ext::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( value ).getAlphabet ( ) );

		ext::map < common::ranked_symbol < DefaultSymbolType, DefaultRankType >, common::SparseBoolVector > vectors;
		for ( std::pair < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > && vector : ext::make_moveable_map ( std::move ( value ).getData ( ) ) )
			vectors.insert ( std::make_pair ( alphabet::SymbolNormalize::normalizeRankedSymbol ( std::move ( vector.first ) ), std::move ( vector.second ) ) );

		return indexes::arbology::NonlinearCompressedBitParallelTreeIndex < > ( std::move ( alphabet ), std::move ( vectors ), std::move ( value ).getJumps ( ), std::move ( value ).getRepeats ( ) );
	}
};

template < class SymbolType, class RankType >
struct xmlApi < indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > > {
	static indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > & data );
};

template < class SymbolType, class RankType >
indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > xmlApi < indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet = alib::xmlApi < ext::set < common::ranked_symbol < SymbolType, RankType > > >::parse ( input );
	ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > data = alib::xmlApi < ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > >::parse ( input );
	ext::vector < int > jumps = alib::xmlApi < ext::vector < int > >::parse ( input );
	ext::vector < unsigned > repeats = alib::xmlApi < ext::vector < unsigned > >::parse ( input );

	indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > res ( std::move ( alphabet ), std::move ( data ), std::move ( jumps ), std::move ( repeats ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType, class RankType >
bool xmlApi < indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType, class RankType >
const std::string & xmlApi < indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > >::xmlTagName ( ) {
	static std::string xmlTagName = "NonlinearCompressedBitParallelTreeIndex";

	return xmlTagName;
}

template < class SymbolType, class RankType >
void xmlApi < indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > >::compose ( ext::deque < sax::Token > & output, const indexes::arbology::NonlinearCompressedBitParallelTreeIndex < SymbolType, RankType > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < ext::set < common::ranked_symbol < SymbolType, RankType > > >::compose ( output, index.getAlphabet ( ) );
	alib::xmlApi < ext::map < common::ranked_symbol < SymbolType, RankType >, common::SparseBoolVector > >::compose ( output, index.getData ( ) );
	alib::xmlApi < ext::vector < int > >::compose ( output, index.getJumps ( ) );
	alib::xmlApi < ext::vector < unsigned > >::compose ( output, index.getRepeats ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace alib */

#endif /* ARBOLOGY_NONLINEAR_COMPRESSED_BIT_PARALLEL_INDEX_H_ */
