/*
 * PostfixRankedTree.h
 *
 *  Created on: May 4, 2017
 *      Author: Aleksandr Shatrovskii
 */

#ifndef POSTFIX_RANKED_TREE_H_
#define POSTFIX_RANKED_TREE_H_

#include <set>
#include <vector>
#include <tree>
#include <sstream>
#include <algorithm>
#include <deque>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include <common/ranked_symbol.hpp>
#include <alphabet/xml/RankedSymbol.h>

#include "../TreeFeatures.h"
#include "../TreeException.h"
#include "../RankedTreeBase.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include <tree/common/TreeNormalize.h>

#include <core/normalize.hpp>

namespace tree {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
template < class SymbolType, class RankType >
class PostfixRankedTree final : public RankedTreeBase, public alib::Components < PostfixRankedTree < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ext::tuple < GeneralAlphabet >, ext::tuple < > > {
	ext::vector < common::ranked_symbol < SymbolType, RankType > > m_Data;

	static ext::vector < common::ranked_symbol < SymbolType, RankType > > toPostfixRanked ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & tree );

public:
	explicit PostfixRankedTree ( ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet, ext::vector < common::ranked_symbol < SymbolType, RankType > > data );
	explicit PostfixRankedTree ( ext::vector < common::ranked_symbol < SymbolType, RankType > > data );
	explicit PostfixRankedTree ( const RankedTree < SymbolType, RankType > & tree );

	virtual RankedTreeBase * clone ( ) const;
	virtual RankedTreeBase * plunder ( ) &&;

	const ext::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < common::ranked_symbol < SymbolType, RankType > > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	void extendAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	void setContent ( ext::vector < common::ranked_symbol < SymbolType, RankType > > data );

	/**
	 * @return List of symbols forming tree (const version).
	 */
	const ext::vector < common::ranked_symbol < SymbolType, RankType > > & getContent ( ) const &;

	ext::vector < common::ranked_symbol < SymbolType, RankType > > && getContent ( ) &&;

	void arityChecksum ( const ext::vector < common::ranked_symbol < SymbolType, RankType > > & data );

	/**
	 * @return true if tree is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PostfixRankedTree & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "PostfixRankedTree";

		return xmlTagName;
	}

	static PostfixRankedTree parse ( ext::deque < sax::Token >::iterator & input );

	static void compose ( ext::deque < sax::Token > & out, const PostfixRankedTree & tree );

	virtual alib::ObjectBase * inc ( ) &&;

	typedef PostfixRankedTree < > normalized_type;
};

} /* namespace tree */

#include "RankedTree.h"

namespace tree {

template < class SymbolType, class RankType >
PostfixRankedTree < SymbolType, RankType >::PostfixRankedTree ( ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet, ext::vector < common::ranked_symbol < SymbolType, RankType > > data ) : alib::Components < PostfixRankedTree, common::ranked_symbol < SymbolType, RankType >, ext::tuple < GeneralAlphabet >, ext::tuple < > > ( ext::make_tuple ( std::move ( alphabet ) ), ext::tuple < > ( ) ) {
	setContent ( std::move ( data ) );
}

template < class SymbolType, class RankType >
PostfixRankedTree < SymbolType, RankType >::PostfixRankedTree ( ext::vector < common::ranked_symbol < SymbolType, RankType > > data ) : PostfixRankedTree ( ext::set < common::ranked_symbol < SymbolType, RankType > > ( data.begin ( ), data.end ( ) ), data ) {
}

template < class SymbolType, class RankType >
PostfixRankedTree < SymbolType, RankType >::PostfixRankedTree ( const RankedTree < SymbolType, RankType > & tree ) : PostfixRankedTree ( tree.getAlphabet ( ), toPostfixRanked ( tree.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
ext::vector < common::ranked_symbol < SymbolType, RankType > > PostfixRankedTree < SymbolType, RankType >::toPostfixRanked ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & tree ) {
	ext::vector < common::ranked_symbol < SymbolType, RankType > > res;

	for ( typename ext::tree < common::ranked_symbol < SymbolType, RankType > >::const_postfix_iterator iter = tree.postfix_begin ( ); iter != tree.postfix_end ( ); ++iter )
		res.push_back ( * iter );

	return res;
}

template < class SymbolType, class RankType >
RankedTreeBase * PostfixRankedTree < SymbolType, RankType >::clone ( ) const {
	return new PostfixRankedTree ( * this );
}

template < class SymbolType, class RankType >
RankedTreeBase * PostfixRankedTree < SymbolType, RankType >::plunder ( ) && {
	return new PostfixRankedTree ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const ext::vector < common::ranked_symbol < SymbolType, RankType > > & PostfixRankedTree < SymbolType, RankType >::getContent ( ) const & {
	return this->m_Data;
}

template < class SymbolType, class RankType >
ext::vector < common::ranked_symbol < SymbolType, RankType > > && PostfixRankedTree < SymbolType, RankType >::getContent ( ) && {
	return std::move ( this->m_Data );
}

template < class SymbolType, class RankType >
void PostfixRankedTree < SymbolType, RankType >::setContent ( ext::vector < common::ranked_symbol < SymbolType, RankType > > data ) {
	arityChecksum ( data );

	ext::set < common::ranked_symbol < SymbolType, RankType > > minimalAlphabet ( data.begin ( ), data.end ( ) );
	ext::set < common::ranked_symbol < SymbolType, RankType > > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet ( ).begin ( ), getAlphabet ( ).end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw TreeException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

template < class SymbolType, class RankType >
void PostfixRankedTree < SymbolType, RankType >::arityChecksum ( const ext::vector < common::ranked_symbol < SymbolType, RankType > > & data ) {
	int arityChecksumAcc = 1;

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : data ) {
		arityChecksumAcc += ( unsigned ) symbol.getRank ( );
		arityChecksumAcc -= 1;
	}

	if ( arityChecksumAcc != 0 ) throw TreeException ( "The string does not form a tree" );
}

template < class SymbolType, class RankType >
bool PostfixRankedTree < SymbolType, RankType >::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

template < class SymbolType, class RankType >
int PostfixRankedTree < SymbolType, RankType >::compare ( const PostfixRankedTree & other ) const {
	auto first = ext::tie ( m_Data, getAlphabet ( ) );
	auto second = ext::tie ( other.m_Data, other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void PostfixRankedTree < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(PostfixRankedTree ";

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

template < class SymbolType, class RankType >
PostfixRankedTree < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : this->m_Data )
		ss << symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

template < class SymbolType, class RankType >
PostfixRankedTree < SymbolType, RankType > PostfixRankedTree < SymbolType, RankType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PostfixRankedTree::getXmlTagName ( ) );
	ext::set < common::ranked_symbol < SymbolType, RankType > > rankedAlphabet = TreeFromXMLParser::parseAlphabet < common::ranked_symbol < SymbolType, RankType > > ( input );
	ext::vector < common::ranked_symbol < SymbolType, RankType > > data = TreeFromXMLParser::parseLinearContent < common::ranked_symbol < SymbolType, RankType > > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PostfixRankedTree::getXmlTagName ( ) );

	return PostfixRankedTree < SymbolType, RankType > ( std::move ( rankedAlphabet ), std::move ( data ) );
}

template < class SymbolType, class RankType >
void PostfixRankedTree < SymbolType, RankType >::compose ( ext::deque < sax::Token > & out, const PostfixRankedTree & tree ) {
	out.emplace_back ( PostfixRankedTree::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, tree.getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, tree.m_Data );
	out.emplace_back ( PostfixRankedTree::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class RankType >
alib::ObjectBase * PostfixRankedTree < SymbolType, RankType >::inc ( ) && {
	return new alib::UniqueObject ( alib::Object ( std::move ( * this ) ), primitive::Integer ( 0 ) );
}

} /* namespace tree */

namespace alib {

template < class SymbolType, class RankType >
class ComponentConstraint < ::tree::PostfixRankedTree < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::PostfixRankedTree < SymbolType, RankType > & tree, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		const ext::vector < common::ranked_symbol < SymbolType, RankType > > & content = tree.getContent ( );

		return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( );
	}

	static bool available ( const ::tree::PostfixRankedTree < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const ::tree::PostfixRankedTree < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
	}

};

template < class SymbolType, class RankType >
struct normalize < tree::PostfixRankedTree < SymbolType, RankType >, typename std::enable_if < ! std::is_same < tree::PostfixRankedTree < SymbolType, RankType >, tree::PostfixRankedTree < > >::value >::type > {
	static tree::PostfixRankedTree < > eval ( tree::PostfixRankedTree < SymbolType, RankType > && value ) {
		ext::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::vector < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > content = alphabet::SymbolNormalize::normalizeRankedSymbols ( std::move ( value ).getContent ( ) );

		return tree::PostfixRankedTree < > ( std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace alib */

#endif /* POSTFIX_RANKED_TREE_H_ */
