/*
 * RankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef RANKED_TREE_H_
#define RANKED_TREE_H_

#include <string>
#include <set>
#include <tree>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include <common/ranked_symbol.hpp>
#include <alphabet/xml/RankedSymbol.h>

#include "../TreeFeatures.h"
#include "../RankedTreeBase.h"
#include "../TreeException.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"

#include <tree/common/TreeNormalize.h>

#include <core/normalize.hpp>

namespace tree {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType, class RankType >
class RankedTree final : public RankedTreeBase, public alib::Components < RankedTree < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ext::tuple < GeneralAlphabet >, ext::tuple < > > {
	ext::tree < common::ranked_symbol < SymbolType, RankType > > m_content;

	void checkAlphabet ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & pattern ) const;

	void checkArities ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & data ) const;

public:
	/**
	 * @copydoc ext::tree < common::ranked_symbol < SymbolType, RankType > >::clone() const
	 */
	virtual RankedTreeBase * clone ( ) const;

	/**
	 * @copydoc ext::tree < common::ranked_symbol < SymbolType, RankType > >::plunder() const
	 */
	virtual RankedTreeBase * plunder ( ) &&;

	explicit RankedTree ( ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet, ext::tree < common::ranked_symbol < SymbolType, RankType > > tree );
	explicit RankedTree ( ext::tree < common::ranked_symbol < SymbolType, RankType > > tree );
	explicit RankedTree ( const UnrankedTree < SymbolType > & other );
	explicit RankedTree ( const PostfixRankedTree < SymbolType, RankType > & other );

	const ext::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const & {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	ext::set < common::ranked_symbol < SymbolType, RankType > > && getAlphabet ( ) && {
		return std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) );
	}

	void extendAlphabet ( const ext::set < common::ranked_symbol < SymbolType, RankType > > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	/**
	 * @return Root node of the regular expression tree
	 */
	const ext::tree < common::ranked_symbol < SymbolType, RankType > > & getContent ( ) const &;

	ext::tree < common::ranked_symbol < SymbolType, RankType > > && getContent ( ) &&;

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( ext::tree < common::ranked_symbol < SymbolType, RankType > > tree );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RankedTree";

		return xmlTagName;
	}

	static RankedTree parse ( ext::deque < sax::Token >::iterator & input );

	static void compose ( ext::deque < sax::Token > & out, const RankedTree & tree );

	virtual alib::ObjectBase * inc ( ) &&;

	typedef RankedTree < > normalized_type;
};

} /* namespace tree */

#include "../unranked/UnrankedTree.h"
#include "PostfixRankedTree.h"

namespace tree {

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::RankedTree ( ext::set < common::ranked_symbol < SymbolType, RankType > > alphabet, ext::tree < common::ranked_symbol < SymbolType, RankType > > tree ) : alib::Components < RankedTree, common::ranked_symbol < SymbolType, RankType >, ext::tuple < GeneralAlphabet >, ext::tuple < > > ( ext::make_tuple ( std::move ( alphabet ) ), ext::tuple < > ( ) ), m_content ( std::move ( tree ) ) {
	checkAlphabet ( m_content );
	checkArities ( m_content );
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::RankedTree ( ext::tree < common::ranked_symbol < SymbolType, RankType > > tree ) : RankedTree ( ext::set < common::ranked_symbol < SymbolType, RankType > > ( tree.prefix_begin ( ), tree.prefix_end ( ) ), tree ) {
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::RankedTree ( const UnrankedTree < SymbolType > & other ) : RankedTree ( TreeAuxiliary::toRankedTree < SymbolType, RankType > ( other.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::RankedTree ( const PostfixRankedTree < SymbolType, RankType > & other) : RankedTree ( TreeAuxiliary::toRankedTree < SymbolType, RankType > ( other.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
RankedTreeBase * RankedTree < SymbolType, RankType >::clone ( ) const {
	return new RankedTree ( * this );
}

template < class SymbolType, class RankType >
RankedTreeBase * RankedTree < SymbolType, RankType >::plunder ( ) && {
	return new RankedTree ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const ext::tree < common::ranked_symbol < SymbolType, RankType > > & RankedTree < SymbolType, RankType >::getContent ( ) const & {
	return m_content;
}

template < class SymbolType, class RankType >
ext::tree < common::ranked_symbol < SymbolType, RankType > > && RankedTree < SymbolType, RankType >::getContent ( ) && {
	return std::move ( m_content );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::checkAlphabet ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & data ) const {
	ext::set < common::ranked_symbol < SymbolType, RankType > > minimalAlphabet ( data.prefix_begin ( ), data.prefix_end ( ) );
	ext::set < common::ranked_symbol < SymbolType, RankType > > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::checkArities ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & data ) const {
	if ( ( size_t ) data.getData ( ).getRank ( ) != data.getChildren ( ).size ( ) )
		throw exception::CommonException ( "Invalid rank." );

	for ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & child : data )
		checkArities ( child );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::setTree ( ext::tree < common::ranked_symbol < SymbolType, RankType > > tree ) {
	checkAlphabet ( tree );
	checkArities ( tree );

	this->m_content = std::move ( tree );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(RankedTree " << this->m_content << ")";
}

template < class SymbolType, class RankType >
int RankedTree < SymbolType, RankType >::compare ( const RankedTree & other ) const {
	auto first = ext::tie ( m_content, getAlphabet() );
	auto second = ext::tie ( other.m_content, other.getAlphabet() );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::nicePrint ( std::ostream & os ) const {
	m_content.nicePrint ( os );
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class RankType >
RankedTree < SymbolType, RankType > RankedTree < SymbolType, RankType >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RankedTree::getXmlTagName() );
	ext::set < common::ranked_symbol < SymbolType, RankType > > rankedAlphabet = TreeFromXMLParser::parseAlphabet < common::ranked_symbol < SymbolType, RankType > > ( input );
	ext::tree < common::ranked_symbol < SymbolType, RankType > > root = TreeFromXMLParser::parseTreeContent < common::ranked_symbol < SymbolType, RankType > > ( input );
	RankedTree < SymbolType, RankType > tree ( std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RankedTree::getXmlTagName() );
	return tree;
}

template < class SymbolType, class RankType >
void RankedTree < SymbolType, RankType >::compose ( ext::deque < sax::Token > & out, const RankedTree & tree ) {
	out.emplace_back ( RankedTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, tree.getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, tree.getContent ( ) );
	out.emplace_back ( RankedTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class RankType >
alib::ObjectBase* RankedTree < SymbolType, RankType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace alib {

template < class SymbolType, class RankType >
class ComponentConstraint< ::tree::RankedTree < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::RankedTree < SymbolType, RankType > & tree, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		const ext::tree < common::ranked_symbol < SymbolType, RankType > > & m_content = tree.getContent ( );
		return std::find(m_content.prefix_begin(), m_content.prefix_end(), symbol) != m_content.prefix_end();
	}

	static bool available ( const ::tree::RankedTree < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const ::tree::RankedTree < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
	}
};

template < class SymbolType, class RankType >
struct normalize < tree::RankedTree < SymbolType, RankType >, typename std::enable_if < ! std::is_same < tree::RankedTree < SymbolType, RankType >, tree::RankedTree < > >::value >::type > {
	static tree::RankedTree < > eval ( tree::RankedTree < SymbolType, RankType > && value ) {
		ext::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( value ).getAlphabet ( ) );
		ext::tree < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > content = tree::TreeNormalize::normalizeRankedTree ( std::move ( value ).getContent ( ) );

		return tree::RankedTree < > ( std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace alib */

#endif /* RANKED_TREE_H_ */
